import socket, threading, time, json
from collections import defaultdict
import loggerric as lr
from enum import Enum

class EventDefaults(Enum):
    CLIENT_CONNECTED = 'client_connected'           # Server
    CLIENT_DISCONNECTED = 'client_disconnected'     # Server
    SERVER_STOPPED = 'server_stopped'               # Client
    UNKNOWN_ID = 'unknown_id'                       # Both

class EventManager:
    def __init__(self):
        self.__handlers = defaultdict(list)
        self.__lock = threading.Lock()
    
    def register(self, event_id:str, func) -> None:
        with self.__lock:
            self.__handlers[event_id].append(func)
            lr.Log.debug(f'Registered handler "{func}" for event "{event_id}"')

    def unregister(self, event_id:str, func) -> None:
        with self.__lock:
            handlers = self.__handlers.get(event_id)
            if handlers and func in handlers:
                handlers.remove(func)
                lr.Log.debug(f'Unregistered handler "{func}" for event "{event_id}"')
    
    def trigger(self, event_id:str, data) -> None:
        handlers = []
        with self.__lock:
            handlers = list(self.__handlers.get(event_id, []))

        if len(handlers) == 0:
            handlers = list(self.__handlers.get(EventDefaults.UNKNOWN_ID, []))

        for func in handlers:
            try:
                func(data)
            except Exception as error:
                lr.Log.error(f'Event handler for "{event_id}" raised: {error}')


import socket, threading, time, json, struct
from sockric._utils import serialize
from sockric._transport import TCP
import loggerric as lr

HEADER_LENGTH_FMT = '!I' # 4 Bytes

def encode(packet_id:str, data, extra_meta:dict=None) -> bytes:
    content_type, payload = serialize(data)

    header = { 'id': packet_id, 'content_type': content_type, 'payload_length': len(payload) }

    if extra_meta:
        header.update(extra_meta)
    
    header_bytes = json.dumps(header, separators=(',', ':'), ensure_ascii=False).encode('utf-8')
    frame = struct.pack(HEADER_LENGTH_FMT, len(header_bytes)) + header_bytes + payload

    return frame

def decode(raw_bytes:bytes) -> tuple[dict, bytes]:
    print('decode')

def decode_from_stream(socket_wrapper:TCP) -> tuple[dict, bytes]:
    header:dict = socket_wrapper.receive(4)
    if header is None:
        return
    
    if header == b'':
        return b''
    
    if len(header) < 4:
        lr.Log.error('Short header length received!')
        return b''
    
    header_length = struct.unpack(HEADER_LENGTH_FMT, header)[0]
    header_bytes = socket_wrapper.receive(header_length)

    if header_bytes is None:
        return
    
    if header_bytes == b'':
        return b''
    
    header = json.loads(header_bytes.decode('utf-8'))
    payload_length = header.get('payload_length', 0)
    payload = b''
    
    if payload_length > 0:
        payload = socket_wrapper.receive(payload_length)
        if payload is None:
            return
        if payload == b'':
            return b''
    
    return header, payload


import socket, threading, time, json, pickle, math
import loggerric as lr

def serialize(data) -> tuple[str, bytes]:
    if isinstance(data, (bytes, bytearray)):
        return 'bytes', bytes(data)

    try:
        j = json.dumps(data, separators=(',', ':'), ensure_ascii=False).encode('utf-8')
        return 'json', j
    except Exception as error:
        lr.Log.debug(f'JSON serialization failed: {error}, faling back to pickle.')

        p = pickle.dumps(data)
        return 'pickle', p

def deserialize(content_type:str, payload_bytes:bytes) -> bytes:
    if content_type == 'bytes':
        return payload_bytes
    
    if content_type == 'json':
        try:
            return json.loads(payload_bytes.decode('utf-8'))
        except Exception as error:
            lr.Log.error(f'Failed JSON decode: {error}')
            return
    
    if content_type == 'pickle':
        try:
            return pickle.loads(payload_bytes)
        except Exception as error:
            lr.Log.error(f'Failed pickle loads: {error}')
            return
        
    lr.Log.warn(f'Unknown content type: "{content_type}"')


import socket, threading, time, json
import loggerric as lr

class TCP:
    def __init__(self, connection:socket.socket, address:tuple[str, int]):
        # Expand parameters scopes
        self.connection = connection
        self.address = address

        self.connection.settimeout(1)

        self.closed = False

        # Thread locks
        self.__recv_lock = threading.Lock()
        self.__send_lock = threading.Lock()
    
    def send(self, raw_bytes:bytes) -> None:
        if self.closed:
            lr.Log.warn('Attempt to send on closed TCP!')
            return
        
        try:
            with self.__send_lock:
                total_sent = 0
                while total_sent < len(raw_bytes):
                    sent = self.connection.send(raw_bytes[total_sent:])

                    if sent == 0:
                        raise RuntimeError('Socket connection broken')

                    total_sent += sent
        except Exception as error:
            lr.Log.error(f'TCP send error to {self.address}: {error}')
            self.close()

    def receive(self, num_bytes:int) -> bytes | None:
        data = bytearray()

        try:
            while len(data) < num_bytes:
                try:
                    chunk = self.connection.recv(num_bytes - len(data))
                except socket.timeout:
                    # No data yet — just return None to let higher-level loop continue
                    return None

                if chunk == b'':
                    # The peer gracefully closed the connection
                    return b''

                data.extend(chunk)

            return bytes(data)
        
        except (ConnectionResetError, OSError) as error:
            lr.Log.warn(f'TCP connection error from {self.address}: {error}')
            self.close()
            return b''
        except Exception as error:
            lr.Log.error(f'Unexpected TCP receive error from {self.address}: {error}')
            self.close()
            return b''

    def close(self) -> None:
        if self.closed:
            return
        
        self.closed = True

        try:
            self.connection.shutdown(socket.SHUT_RDWR)
        except:
            pass

        try:
            self.connection.close()
        except:
            pass


from sockric._events import EventDefaults
from sockric._transport import TCP
from sockric._server import Server
from sockric._client import Client

# Expose these functions/classes
__all__ = ['Server', 'TCP', 'EventDefaults', 'Client']


from sockric._packet import encode, decode, decode_from_stream
from sockric._events import EventManager, EventDefaults
from sockric._utils import serialize, deserialize
from sockric._transport import TCP
import socket, threading, time, json
import loggerric as lr

class Client:
    def __init__(self, host_ip:str, port:int):
        self.__host_ip = host_ip
        self.__port = port

        self.password:str = None
        self.is_connected = False
        self.__events = EventManager()
    
        self.__socket:socket.socket = None
        self.__tcp_wrapper:TCP = None
        self.__recv_thread:threading.Thread = None

    def get_host(self) -> tuple[str, int]:
        return self.__host_ip, self.__port

    def connect(self, password:str=None) -> None:
        if self.is_connected:
            lr.Log.warn('Client is already connected!')
            return
    
        self.password = password

        lr.Log.debug(f'Client connecting to {self.__host_ip}:{self.__port}')

        self.__socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__socket.settimeout(2.0)

        try:
            self.__socket.connect((self.__host_ip, self.__port))
        except Exception as error:
            lr.Log.error(f'TCP connection failed: {error}')
            return

        self.__tcp_wrapper = TCP(self.__socket, (self.__host_ip, self.__port))

        self.is_connected = True

        self.__recv_thread = threading.Thread(target=self.__tcp_recv_loop, daemon=True)
        self.__recv_thread.start()

        lr.Log.info('Client successfully connected (awaiting handshake confirmation)')

        handshake = { 'password': self.password }
        frame = encode('__handshake__', handshake)
        try:
            self.__tcp_wrapper.send(frame)
        except Exception as error:
            lr.Log.error(f'Handshake send failed: {error}')
    
    def send(self, packet_id:str, data):
        if not self.is_connected:
            lr.Log.warn('Cannot send while client not connected!')
            return
        
        frame = encode(packet_id, data)

        try:
            self.__tcp_wrapper.send(frame)
        except Exception as error:
            lr.Log.error(f'TCP send error: {error}')

    def disconnect(self):
        if not self.is_connected:
            lr.Log.warn("Client isn't connected!")
            return

        lr.Log.info("Client disconnecting!")
        self.is_connected = False

        try:
            if self.__tcp_wrapper and not self.__tcp_wrapper.closed:
                # Send a disconnect packet first
                try:
                    frame = encode('__disconnect__', None)
                    self.__tcp_wrapper.send(frame)
                    time.sleep(0.1)  # Give server time to process
                except Exception:
                    pass
                self.__tcp_wrapper.close()
        except Exception as e:
            lr.Log.warn(f"Error closing TCP: {e}")

        if self.__recv_thread and self.__recv_thread.is_alive():
            self.__recv_thread.join(timeout=1.0)

    def on_packet(self, id:str) -> None:
        def decorator(func):
            self.__events.register(id, func)
            return func
        
        return decorator

    def __tcp_recv_loop(self) -> None:
        lr.Log.debug('Client TCP recv loop started!')

        while self.is_connected:
            try:
                response = decode_from_stream(self.__tcp_wrapper)
                if response is None:
                    continue
                    
                if response == b'':
                    lr.Log.info('Server closed TCP connection')
                    break

                header, payload = response

                # Check if this is the handshake response
                packet_id = header.get('id')
                if packet_id == '__auth_success__':
                    lr.Log.info('Handshake authentication passed!')
                elif packet_id == '__auth_failed__':
                    lr.Log.info('Handshake authentication failed!')
                    break

                data = deserialize(header.get('content_type'), payload)
                self.__events.trigger(packet_id, { 'data': data, 'header': header })
            except Exception as error:
                lr.Log.error(f'Client TCP recv loop error: {error}')
                break
        
        self.disconnect()


from sockric._packet import encode, decode, decode_from_stream
from sockric._events import EventManager, EventDefaults
from sockric._utils import serialize, deserialize
from sockric._transport import TCP
import socket, threading, time, json
import loggerric as lr

class Server:
    def __init__(self, host_ip:str, port:int, password:str=None):
        # Expand parameters scopes
        self.__host_ip = host_ip
        self.__port = port
        self.password = password
    
        self.__is_running = False
        self.__client_id_counter = 0
        self.__clients:dict[int, dict] = {}
        self.__events = EventManager()

        self.__tcp_socket:socket.socket = None
        self.__clients_lock:threading.Lock = None
        self.__recv_threads:list[threading.Thread] = None
        self.__accept_thread:threading.Thread = None

    def get_host(self) -> tuple[str, int]:
        return self.__host_ip, self.__port

    def start(self) -> None:
        # Check if the server is already running
        if self.__is_running:
            lr.Log.warn('Server is already running!')
            return
        
        lr.Log.debug(f'Starting server, hosting on: {self.__host_ip}:{self.__port}')

        # Set up the TCP socket as a server
        self.__tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__tcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.__tcp_socket.bind((self.__host_ip, self.__port))
        self.__tcp_socket.listen(8) # Queue size

        self.__is_running = True
        self.__clients_lock = threading.Lock()
        self.__recv_threads:dict[int, threading.Thread] = {}

        # Set up the client accept thread
        self.__accept_thread = threading.Thread(target=self.__client_accept_loop, daemon=True)
        self.__accept_thread.start()

        lr.Log.info(f'Server successfully started, hosting on ({self.__host_ip}:{self.__port})')

    def stop(self) -> None:
        # Check if the server is running
        if not self.__is_running:
            lr.Log.warn('Server is not running!')
            return

        self.broadcast(EventDefaults.SERVER_STOPPED, '__server_stopped__')

        self.__is_running = False

        lr.Log.debug('Stopping server!')

    def send(self, packet_id:str, client_ids:int|list, data) -> None:
        if isinstance(client_ids, int):
            client_ids = [client_ids]
        
        frame = encode(packet_id, data)
        with self.__clients_lock:
            for client_id, info in self.__clients.items():
                if client_ids and client_id not in client_ids:
                    continue

                try:
                    tcp:TCP = info.get('tcp')
                    tcp.send(frame)
                except Exception as error:
                    lr.Log.error(f'Failed to send to client {client_id}: {error}')

    def broadcast(self, packet_id:str, data, exclude_clients:list[int]=None) -> None:
        if exclude_clients is None:
            exclude_clients = []
        
        with self.__clients_lock:
            for client_id in list(self.__clients.keys()):
                if client_id in exclude_clients:
                    continue

                self.send(packet_id, client_id, data)

    def on_packet(self, id:str):
        def decorator(func):
            self.__events.register(id, func)
            return func
        
        return decorator

    def __client_accept_loop(self) -> None:
        lr.Log.debug('Client accepting loop started!')

        def attempt_accecpt():
            connection, address = self.__tcp_socket.accept()
            lr.Log.debug(f'Incomming TCP connection from: {address}')

            tcp_wrapper = TCP(connection, address)

            response:tuple[dict, dict] = None
            try:
                response = decode_from_stream(tcp_wrapper)
            except Exception as error:
                lr.Log.warn(f'Handshake decode error: {error}')
            
            if response == b'':
                lr.Log.warn('Connection closed by peer during handshake!')
                tcp_wrapper.close()
                return
            else:
                header, payload = response
                meta:dict = deserialize(header.get('content_type'), payload)

                password = meta.get('password')
                if header.get('id') == '__handshake__':
                    if self.password and self.password != password:
                        lr.Log.info(f'Rejecting connection from {address}: "{password}" Incorrect password!')

                        failure = encode('__auth_failed__', { 'reason': 'bad_password' })
                        tcp_wrapper.send(failure)
                        tcp_wrapper.close()
                        return
                    else:
                        success = encode('__auth_success__', { 'reason': ('good_password' if self.password else 'no_password_set') })
                        tcp_wrapper.send(success)
            
            with self.__clients_lock:
                client_id = self.__client_id_counter
                self.__client_id_counter += 1
                self.__clients[client_id] = { 'tcp': tcp_wrapper, 'address': address }
            
            lr.Log.info(f'Client {client_id} connected from: {address}')

            self.__events.trigger(EventDefaults.CLIENT_CONNECTED, { 'client_id': client_id, 'address': address })

            thread = threading.Thread(target=self.__tcp_client_loop, args=(client_id,), daemon=True)
            thread.start()
            self.__recv_threads[client_id] = thread
        while self.__is_running:
            try:
                attempt_accecpt()
            except Exception as error:
                if self.__is_running:
                    lr.Log.error(f'Accept loop error: {error}')

    def __tcp_client_loop(self, client_id:int) -> None:
        lr.Log.debug(f'TCP recv loop for client {client_id} started!')

        tcp_wrapper = None
        with self.__clients_lock:
            info = self.__clients.get(client_id)
            if not info:
                return
            
            tcp_wrapper = info.get('tcp')
        
        while self.__is_running:
            try:
                response = decode_from_stream(tcp_wrapper)
                if response is None:
                    continue

                if response == b'':
                    lr.Log.debug(f'Client {client_id} disconnected')
                    break

                header, payload = response
                packet_id = header.get('id')
                
                # Handle graceful disconnect packet
                if packet_id == '__disconnect__':
                    lr.Log.debug(f'Client {client_id} sent disconnect packet')
                    break

                header, payload = response
                data = deserialize(header.get('content_type'), payload)
                self.__events.trigger(header.get('id'), { 'client_id': client_id, 'data': data, 'header': header })
            except Exception as error:
                lr.Log.error(f'Error in TCP recv loop for client {client_id}: {error}')
                break
        
        self.__cleanup_client(client_id)
    
    def __cleanup_client(self, client_id:int) -> None:
        with self.__clients_lock:
            client_info = self.__clients.pop(client_id, None)
            client_thread = self.__recv_threads.pop(client_id, None)
        
        if not client_info:
            lr.Log.warn(f'Attempted to clean up non-existent client {client_id}')
            return
            
        # Trigger disconnect event with client info
        self.__events.trigger(EventDefaults.CLIENT_DISCONNECTED, {'client_id': client_id, 'address': client_info.get('tcp').address})
        
        if client_thread and threading.current_thread() != client_thread:
            try:
                client_thread.join(1.0)
            except Exception as error:
                lr.Log.error(f'Error while joining client {client_id} thread: {error}')

        tcp_wrapper:TCP = client_info.get('tcp')
        if tcp_wrapper:
            try:
                tcp_wrapper.close()
                lr.Log.info(f'Closed TCP socket for client {client_id}')
            except Exception as error:
                lr.Log.warn(f'Error closing TCP socket for client {client_id}: {error}')